/************************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

********************************************************/

/* $XConsortium: swaprep.c,v 1.36 89/07/03 19:50:49 rws Exp $ */

#include "X.h"
#define NEED_REPLIES
#define NEED_EVENTS
#include "Xproto.h"
#include "misc.h"
#include "dixstruct.h"
#include "fontstruct.h"
#include "scrnintstr.h"

void SwapVisual(), SwapConnSetup(), SwapWinRoot();

/* Thanks to Jack Palevich for testing and subsequently rewriting all this */
void
Swap32Write(pClient, size, pbuf)
    ClientPtr	pClient;
    int		size;  /* in bytes */
    long *pbuf;
{
    int		i;

    size >>= 2;
    for(i = 0; i < size; i++)
    /* brackets are mandatory here, because "swapl" macro expands
       to several statements */
    {   
	pbuf[i] = lswapl(pbuf[i]);
    }
    (void)WriteToClient(pClient, size << 2, (char *) pbuf);
}

void
CopySwap32Write(pClient, size, pbuf)
    ClientPtr	pClient;
    int		size;   /* in bytes */
    long	*pbuf;
{
    int bufsize = size;
    long *pbufT;
    long *from, *to, *fromLast, *toLast;
    long tmpbuf[1];
    
    /* Allocate as big a buffer as we can... */
    while (!(pbufT = (long *) ALLOCATE_LOCAL(bufsize)))
    {
        bufsize >>= 1;
	if (bufsize == 4)
	{
	    pbufT = tmpbuf;
	    break;
	}
    }
    
    /* convert lengths from # of bytes to # of longs */
    size >>= 2;
    bufsize >>= 2;

    from = pbuf;
    fromLast = from + size;
    while (from < fromLast) {
	int nbytes;
        to = pbufT;
        toLast = to + min (bufsize, fromLast - from);
        nbytes = (toLast - to) << 2;
        while (to < toLast) {
            /* can't write "cpswapl(*from++, *to++)" because cpswapl is a macro
	       that evaulates its args more than once */
	    cpswapl(*from, *to);
            from++;
            to++;
	    }
	(void)WriteToClient (pClient, nbytes, (char *) pbufT);
	}

    if (pbufT != tmpbuf)
	DEALLOCATE_LOCAL ((char *) pbufT);
}

void
CopySwap16Write(pClient, size, pbuf)
    ClientPtr	pClient;
    int		size;   /* in bytes */
    short	*pbuf;
{
    int bufsize = size;
    short *pbufT;
    short *from, *to, *fromLast, *toLast;
    short tmpbuf[2];
    
    /* Allocate as big a buffer as we can... */
    while (!(pbufT = (short *) ALLOCATE_LOCAL(bufsize)))
    {
        bufsize >>= 1;
	if (bufsize == 4)
	{
	    pbufT = tmpbuf;
	    break;
	}
    }
    
    /* convert lengths from # of bytes to # of shorts */
    size >>= 1;
    bufsize >>= 1;

    from = pbuf;
    fromLast = from + size;
    while (from < fromLast) {
	int nbytes;
        to = pbufT;
        toLast = to + min (bufsize, fromLast - from);
        nbytes = (toLast - to) << 1;
        while (to < toLast) {
            /* can't write "cpswaps(*from++, *to++)" because cpswaps is a macro
	       that evaulates its args more than once */
	    cpswaps(*from, *to);
            from++;
            to++;
	    }
	(void)WriteToClient (pClient, nbytes, (char *) pbufT);
	}

    if (pbufT != tmpbuf)
	DEALLOCATE_LOCAL ((char *) pbufT);
}


/* Extra-small reply */
void
SGenericReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGenericReply		*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

/* Extra-large reply */
void
SGetWindowAttributesReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetWindowAttributesReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->visualID = lswapl(pRep->visualID);
    pRep->class = lswaps(pRep->class);
    pRep->backingBitPlanes = lswapl(pRep->backingBitPlanes);
    pRep->backingPixel = lswapl(pRep->backingPixel);
    pRep->colormap = lswapl(pRep->colormap);
    pRep->allEventMasks = lswapl(pRep->allEventMasks);
    pRep->yourEventMask = lswapl(pRep->yourEventMask);
    pRep->doNotPropagateMask = lswaps(pRep->doNotPropagateMask);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetGeometryReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xGetGeometryReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->root = lswapl(pRep->root);
    pRep->x = lswaps(pRep->x);
    pRep->y = lswaps(pRep->y);
    pRep->width = lswaps(pRep->width);
    pRep->height = lswaps(pRep->height);
    pRep->borderWidth = lswaps(pRep->borderWidth);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SQueryTreeReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xQueryTreeReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->root = lswapl(pRep->root);
    pRep->parent = lswapl(pRep->parent);
    pRep->nChildren = lswaps(pRep->nChildren);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SInternAtomReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xInternAtomReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->atom = lswapl(pRep->atom);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetAtomNameReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetAtomNameReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nameLength = lswaps(pRep->nameLength);
    (void)WriteToClient(pClient, size, (char *) pRep);
}


void
SGetPropertyReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetPropertyReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->propertyType = lswapl(pRep->propertyType);
    pRep->bytesAfter = lswapl(pRep->bytesAfter);
    pRep->nItems = lswapl(pRep->nItems);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SListPropertiesReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xListPropertiesReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nProperties = lswaps(pRep->nProperties);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetSelectionOwnerReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetSelectionOwnerReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->owner = lswapl(pRep->owner);
    (void)WriteToClient(pClient, size, (char *) pRep);
}


void
SQueryPointerReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xQueryPointerReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->root = lswapl(pRep->root);
    pRep->child = lswapl(pRep->child);
    pRep->rootX = lswaps(pRep->rootX);
    pRep->rootY = lswaps(pRep->rootY);
    pRep->winX = lswaps(pRep->winX);
    pRep->winY = lswaps(pRep->winY);
    pRep->mask = lswaps(pRep->mask);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SwapTimeCoordWrite(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xTimecoord			*pRep;
{
    int	i, n;
    xTimecoord			*pRepT;

    n = size / sizeof(xTimecoord);
    pRepT = pRep;
    for(i = 0; i < n; i++)
    {
	SwapTimecoord(pRepT);
	pRepT++;
    }
    (void)WriteToClient(pClient, size, (char *) pRep);

}
void
SGetMotionEventsReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetMotionEventsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nEvents = lswapl(pRep->nEvents);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
STranslateCoordsReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xTranslateCoordsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->child = lswapl(pRep->child);
    pRep->dstX = lswaps(pRep->dstX);
    pRep->dstY = lswaps(pRep->dstY);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetInputFocusReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xGetInputFocusReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->focus = lswapl(pRep->focus);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

/* extra long reply */
void
SQueryKeymapReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xQueryKeymapReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

static void
SwapCharInfo(pInfo)
    xCharInfo	*pInfo;
{

    pInfo->leftSideBearing = lswaps(pInfo->leftSideBearing);
    pInfo->rightSideBearing = lswaps(pInfo->rightSideBearing);
    pInfo->characterWidth = lswaps(pInfo->characterWidth);
    pInfo->ascent = lswaps(pInfo->ascent);
    pInfo->descent = lswaps(pInfo->descent);
    pInfo->attributes = lswaps(pInfo->attributes);
}

static void
SwapFontInfo(pr)
    xQueryFontReply *pr;
{

    pr->minCharOrByte2 = lswaps(pr->minCharOrByte2);
    pr->maxCharOrByte2 = lswaps(pr->maxCharOrByte2);
    pr->defaultChar = lswaps(pr->defaultChar);
    pr->nFontProps = lswaps(pr->nFontProps);
    pr->fontAscent = lswaps(pr->fontAscent);
    pr->fontDescent = lswaps(pr->fontDescent);
    SwapCharInfo( &pr->minBounds);
    SwapCharInfo( &pr->maxBounds);
    pr->nCharInfos = lswapl(pr->nCharInfos);
}

static void
SwapFont( pr, hasGlyphs)
    xQueryFontReply *	pr;
    Bool hasGlyphs;
{
    unsigned	i;
    xCharInfo *	pxci;
    unsigned	nchars, nprops;
    u_long	*pby;

    pr->sequenceNumber = lswaps(pr->sequenceNumber);
    pr->length = lswapl(pr->length);
    nchars = pr->nCharInfos;
    nprops = pr->nFontProps;
    SwapFontInfo(pr);
    pby = (u_long *) &pr[1];
    /* Font properties are an atom and either an int32 or a CARD32, so
     * they are always 2 4 byte values */
    for(i = 0; i < nprops; i++)
    {
	*pby = lswapl(*pby);
	pby++;
	*pby = lswapl(*pby);
	pby++;
    }
    if (hasGlyphs)
    {
	pxci = (xCharInfo *)pby;
	for(i = 0; i< nchars; i++, pxci++)
	    SwapCharInfo(pxci);
    }
}

void
SQueryFontReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xQueryFontReply	*pRep;
{
    SwapFont(pRep, TRUE);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SQueryTextExtentsReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xQueryTextExtentsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->fontAscent = lswaps(pRep->fontAscent);
    pRep->fontDescent = lswaps(pRep->fontDescent);
    pRep->overallAscent = lswaps(pRep->overallAscent);
    pRep->overallDescent = lswaps(pRep->overallDescent);
    pRep->overallWidth = lswapl(pRep->overallWidth);
    pRep->overallLeft = lswapl(pRep->overallLeft);
    pRep->overallRight = lswapl(pRep->overallRight);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SListFontsReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xListFontsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nFonts = lswaps(pRep->nFonts);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SListFontsWithInfoReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xListFontsWithInfoReply	*pRep;
{
    SwapFont((xQueryFontReply *)pRep, FALSE);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetFontPathReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xGetFontPathReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nPaths = lswaps(pRep->nPaths);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetImageReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xGetImageReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->visual = lswapl(pRep->visual);
    (void)WriteToClient(pClient, size, (char *) pRep);
    /* Fortunately, image doesn't need swapping */
}

void
SListInstalledColormapsReply(pClient, size, pRep)
    ClientPtr				pClient;
    int					size;
    xListInstalledColormapsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nColormaps = lswaps(pRep->nColormaps);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SAllocColorReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xAllocColorReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->red = lswaps(pRep->red);
    pRep->green = lswaps(pRep->green);
    pRep->blue = lswaps(pRep->blue);
    pRep->pixel = lswapl(pRep->pixel);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SAllocNamedColorReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xAllocNamedColorReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->pixel = lswapl(pRep->pixel);
    pRep->exactRed = lswaps(pRep->exactRed);
    pRep->exactGreen = lswaps(pRep->exactGreen);
    pRep->exactBlue = lswaps(pRep->exactBlue);
    pRep->screenRed = lswaps(pRep->screenRed);
    pRep->screenGreen = lswaps(pRep->screenGreen);
    pRep->screenBlue = lswaps(pRep->screenBlue);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SAllocColorCellsReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xAllocColorCellsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nPixels = lswaps(pRep->nPixels);
    pRep->nMasks = lswaps(pRep->nMasks);
    (void)WriteToClient(pClient, size, (char *) pRep);
}


void
SAllocColorPlanesReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xAllocColorPlanesReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nPixels = lswaps(pRep->nPixels);
    pRep->redMask = lswapl(pRep->redMask);
    pRep->greenMask = lswapl(pRep->greenMask);
    pRep->blueMask = lswapl(pRep->blueMask);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SQColorsExtend(pClient, size, prgb)
    ClientPtr	pClient;
    int		size;
    xrgb	*prgb;
{
    int		i, n;
    xrgb	*prgbT;

    n = size / sizeof(xrgb);
    prgbT = prgb;
    for(i = 0; i < n; i++)
    {
	SwapRGB(prgbT);
	prgbT++;
    }
    (void)WriteToClient(pClient, size, (char *) prgb);
}

void
SQueryColorsReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xQueryColorsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nColors = lswaps(pRep->nColors);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SLookupColorReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xLookupColorReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->exactRed = lswaps(pRep->exactRed);
    pRep->exactGreen = lswaps(pRep->exactGreen);
    pRep->exactBlue = lswaps(pRep->exactBlue);
    pRep->screenRed = lswaps(pRep->screenRed);
    pRep->screenGreen = lswaps(pRep->screenGreen);
    pRep->screenBlue = lswaps(pRep->screenBlue);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SQueryBestSizeReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xQueryBestSizeReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->width = lswaps(pRep->width);
    pRep->height = lswaps(pRep->height);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SListExtensionsReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xListExtensionsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetKeyboardMappingReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetKeyboardMappingReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetPointerMappingReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetPointerMappingReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetModifierMappingReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetModifierMappingReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetKeyboardControlReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetKeyboardControlReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->ledMask = lswapl(pRep->ledMask);
    pRep->bellPitch = lswaps(pRep->bellPitch);
    pRep->bellDuration = lswaps(pRep->bellDuration);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetPointerControlReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetPointerControlReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->accelNumerator = lswaps(pRep->accelNumerator);
    pRep->accelDenominator = lswaps(pRep->accelDenominator);
    pRep->threshold = lswaps(pRep->threshold);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SGetScreenSaverReply(pClient, size, pRep)
    ClientPtr			pClient;
    int				size;
    xGetScreenSaverReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->timeout = lswaps(pRep->timeout);
    pRep->interval = lswaps(pRep->interval);
    (void)WriteToClient(pClient, size, (char *) pRep);
}

void
SLHostsExtend(pClient, size, buf)
    ClientPtr		pClient;
    int			size;
    char		*buf;
{
    char *bufT = buf;
    char *endbuf = buf + size;
    while (bufT < endbuf) {
	xHostEntry *host = (xHostEntry *) bufT;
	int len = host->length;
	host->length = lswaps(len);
	bufT += sizeof (xHostEntry) + (((len + 3) >> 2) << 2);
	}
    (void)WriteToClient (pClient, size, buf);
}

void
SListHostsReply(pClient, size, pRep)
    ClientPtr		pClient;
    int			size;
    xListHostsReply	*pRep;
{

    pRep->sequenceNumber = lswaps(pRep->sequenceNumber);
    pRep->length = lswapl(pRep->length);
    pRep->nHosts = lswaps(pRep->nHosts);
    (void)WriteToClient(pClient, size, (char *) pRep);
}



void
SErrorEvent(from, to)
    xError	*from, *to;
{
    to->type = X_Error;
    to->errorCode = from->errorCode;
    cpswaps(from->sequenceNumber, to->sequenceNumber);
    cpswapl(from->resourceID, to->resourceID);
    cpswaps(from->minorCode, to->minorCode);
    to->majorCode = from->majorCode;
}

void
SKeyButtonPtrEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    to->u.u.detail = from->u.u.detail;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.keyButtonPointer.time,
        to->u.keyButtonPointer.time);
    cpswapl(from->u.keyButtonPointer.root,
        to->u.keyButtonPointer.root);
    cpswapl(from->u.keyButtonPointer.event,
        to->u.keyButtonPointer.event);
    cpswapl(from->u.keyButtonPointer.child,
        to->u.keyButtonPointer.child);
    cpswaps(from->u.keyButtonPointer.rootX,
        to->u.keyButtonPointer.rootX);
    cpswaps(from->u.keyButtonPointer.rootY,
	to->u.keyButtonPointer.rootY);
    cpswaps(from->u.keyButtonPointer.eventX,
        to->u.keyButtonPointer.eventX);
    cpswaps(from->u.keyButtonPointer.eventY,
        to->u.keyButtonPointer.eventY);
    cpswaps(from->u.keyButtonPointer.state,
        to->u.keyButtonPointer.state);
    to->u.keyButtonPointer.sameScreen = 
	from->u.keyButtonPointer.sameScreen;
}

void
SEnterLeaveEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    to->u.u.detail = from->u.u.detail;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.enterLeave.time, to->u.enterLeave.time);
    cpswapl(from->u.enterLeave.root, to->u.enterLeave.root);
    cpswapl(from->u.enterLeave.event, to->u.enterLeave.event);
    cpswapl(from->u.enterLeave.child, to->u.enterLeave.child);
    cpswaps(from->u.enterLeave.rootX, to->u.enterLeave.rootX);
    cpswaps(from->u.enterLeave.rootY, to->u.enterLeave.rootY);
    cpswaps(from->u.enterLeave.eventX, to->u.enterLeave.eventX);
    cpswaps(from->u.enterLeave.eventY, to->u.enterLeave.eventY);
    cpswaps(from->u.enterLeave.state, to->u.enterLeave.state);
    to->u.enterLeave.mode = from->u.enterLeave.mode;
    to->u.enterLeave.flags = from->u.enterLeave.flags;
}

void
SFocusEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    to->u.u.detail = from->u.u.detail;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.focus.window, to->u.focus.window);
    to->u.focus.mode = from->u.focus.mode;
}

void
SExposeEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.expose.window, to->u.expose.window);
    cpswaps(from->u.expose.x, to->u.expose.x);
    cpswaps(from->u.expose.y, to->u.expose.y);
    cpswaps(from->u.expose.width, to->u.expose.width);
    cpswaps(from->u.expose.height, to->u.expose.height);
    cpswaps(from->u.expose.count, to->u.expose.count);
}

void
SGraphicsExposureEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.graphicsExposure.drawable,
        to->u.graphicsExposure.drawable);
    cpswaps(from->u.graphicsExposure.x, 
	to->u.graphicsExposure.x);
    cpswaps(from->u.graphicsExposure.y, 
	to->u.graphicsExposure.y);
    cpswaps(from->u.graphicsExposure.width, 
	to->u.graphicsExposure.width);
    cpswaps(from->u.graphicsExposure.height, 
	to->u.graphicsExposure.height);
    cpswaps(from->u.graphicsExposure.minorEvent,
        to->u.graphicsExposure.minorEvent);
    cpswaps(from->u.graphicsExposure.count,
	to->u.graphicsExposure.count);
    to->u.graphicsExposure.majorEvent = 
    	from->u.graphicsExposure.majorEvent;
}

void
SNoExposureEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.noExposure.drawable, to->u.noExposure.drawable);
    cpswaps(from->u.noExposure.minorEvent, to->u.noExposure.minorEvent);
    to->u.noExposure.majorEvent = from->u.noExposure.majorEvent;
}

void
SVisibilityEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.visibility.window, to->u.visibility.window);
    to->u.visibility.state = from->u.visibility.state;
}

void
SCreateNotifyEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.createNotify.window, to->u.createNotify.window);
    cpswapl(from->u.createNotify.parent, to->u.createNotify.parent);
    cpswaps(from->u.createNotify.x, to->u.createNotify.x);
    cpswaps(from->u.createNotify.y, to->u.createNotify.y);
    cpswaps(from->u.createNotify.width, to->u.createNotify.width);
    cpswaps(from->u.createNotify.height, to->u.createNotify.height);
    cpswaps(from->u.createNotify.borderWidth,
        to->u.createNotify.borderWidth);
    to->u.createNotify.override = from->u.createNotify.override;
}

void
SDestroyNotifyEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.destroyNotify.event, to->u.destroyNotify.event);
    cpswapl(from->u.destroyNotify.window, to->u.destroyNotify.window);
}

void
SUnmapNotifyEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.unmapNotify.event, to->u.unmapNotify.event);
    cpswapl(from->u.unmapNotify.window, to->u.unmapNotify.window);
    to->u.unmapNotify.fromConfigure = from->u.unmapNotify.fromConfigure;
}

void
SMapNotifyEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.mapNotify.event, to->u.mapNotify.event);
    cpswapl(from->u.mapNotify.window, to->u.mapNotify.window);
    to->u.mapNotify.override = from->u.mapNotify.override;
}

void
SMapRequestEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.mapRequest.parent, to->u.mapRequest.parent);
    cpswapl(from->u.mapRequest.window, to->u.mapRequest.window);
}

void
SReparentEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.reparent.event, to->u.reparent.event);
    cpswapl(from->u.reparent.window, to->u.reparent.window);
    cpswapl(from->u.reparent.parent, to->u.reparent.parent);
    cpswaps(from->u.reparent.x, to->u.reparent.x);
    cpswaps(from->u.reparent.y, to->u.reparent.y);
    to->u.reparent.override = from->u.reparent.override;
}

void
SConfigureNotifyEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.configureNotify.event,
        to->u.configureNotify.event);
    cpswapl(from->u.configureNotify.window,
        to->u.configureNotify.window);
    cpswapl(from->u.configureNotify.aboveSibling,
        to->u.configureNotify.aboveSibling);
    cpswaps(from->u.configureNotify.x, to->u.configureNotify.x);
    cpswaps(from->u.configureNotify.y, to->u.configureNotify.y);
    cpswaps(from->u.configureNotify.width, to->u.configureNotify.width);
    cpswaps(from->u.configureNotify.height,
        to->u.configureNotify.height);
    cpswaps(from->u.configureNotify.borderWidth,
        to->u.configureNotify.borderWidth);
    to->u.configureNotify.override = from->u.configureNotify.override;
}

void
SConfigureRequestEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    to->u.u.detail = from->u.u.detail;  /* actually stack-mode */
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.configureRequest.parent,
        to->u.configureRequest.parent);
    cpswapl(from->u.configureRequest.window,
        to->u.configureRequest.window);
    cpswapl(from->u.configureRequest.sibling,
        to->u.configureRequest.sibling);
    cpswaps(from->u.configureRequest.x, to->u.configureRequest.x);
    cpswaps(from->u.configureRequest.y, to->u.configureRequest.y);
    cpswaps(from->u.configureRequest.width,
        to->u.configureRequest.width);
    cpswaps(from->u.configureRequest.height,
        to->u.configureRequest.height);
    cpswaps(from->u.configureRequest.borderWidth,
        to->u.configureRequest.borderWidth);
    cpswaps(from->u.configureRequest.valueMask,
        to->u.configureRequest.valueMask);
}


void
SGravityEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.gravity.event, to->u.gravity.event);
    cpswapl(from->u.gravity.window, to->u.gravity.window);
    cpswaps(from->u.gravity.x, to->u.gravity.x);
    cpswaps(from->u.gravity.y, to->u.gravity.y);
}

void
SResizeRequestEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.resizeRequest.window, to->u.resizeRequest.window);
    cpswaps(from->u.resizeRequest.width, to->u.resizeRequest.width);
    cpswaps(from->u.resizeRequest.height, to->u.resizeRequest.height);
}

void
SCirculateEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    to->u.u.detail = from->u.u.detail;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.circulate.event, to->u.circulate.event);
    cpswapl(from->u.circulate.window, to->u.circulate.window);
    cpswapl(from->u.circulate.parent, to->u.circulate.parent);
    to->u.circulate.place = from->u.circulate.place;
}

void
SPropertyEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.property.window, to->u.property.window);
    cpswapl(from->u.property.atom, to->u.property.atom);
    cpswapl(from->u.property.time, to->u.property.time);
    to->u.property.state = from->u.property.state;
}

void
SSelectionClearEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.selectionClear.time, to->u.selectionClear.time);
    cpswapl(from->u.selectionClear.window, to->u.selectionClear.window);
    cpswapl(from->u.selectionClear.atom, to->u.selectionClear.atom);
}

void
SSelectionRequestEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.selectionRequest.time, to->u.selectionRequest.time);
    cpswapl(from->u.selectionRequest.owner,
        to->u.selectionRequest.owner);
    cpswapl(from->u.selectionRequest.requestor,
	to->u.selectionRequest.requestor);
    cpswapl(from->u.selectionRequest.selection,
	to->u.selectionRequest.selection);
    cpswapl(from->u.selectionRequest.target,
        to->u.selectionRequest.target);
    cpswapl(from->u.selectionRequest.property,
	to->u.selectionRequest.property);
}

void
SSelectionNotifyEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.selectionNotify.time, to->u.selectionNotify.time);
    cpswapl(from->u.selectionNotify.requestor,
	to->u.selectionNotify.requestor);
    cpswapl(from->u.selectionNotify.selection,
	to->u.selectionNotify.selection);
    cpswapl(from->u.selectionNotify.target,
	to->u.selectionNotify.target);
    cpswapl(from->u.selectionNotify.property,
        to->u.selectionNotify.property);
}

void
SColormapEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.colormap.window, to->u.colormap.window);
    cpswapl(from->u.colormap.colormap, to->u.colormap.colormap);
    to->u.colormap.new = from->u.colormap.new;
    to->u.colormap.state = from->u.colormap.state;
}

void
SMappingEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    to->u.mappingNotify.request = from->u.mappingNotify.request;
    to->u.mappingNotify.firstKeyCode =
	from->u.mappingNotify.firstKeyCode;
    to->u.mappingNotify.count = from->u.mappingNotify.count;
}

void
SClientMessageEvent(from, to)
    xEvent	*from, *to;
{
    to->u.u.type = from->u.u.type;
    to->u.u.detail = from->u.u.detail;  /* actually format */
    cpswaps(from->u.u.sequenceNumber, to->u.u.sequenceNumber);
    cpswapl(from->u.clientMessage.window, to->u.clientMessage.window);
    cpswapl(from->u.clientMessage.u.l.type, 
	    to->u.clientMessage.u.l.type);
    switch (from->u.u.detail) {
       case 8:
          bcopy(from->u.clientMessage.u.b.bytes,
		to->u.clientMessage.u.b.bytes, 20);
	  break;
       case 16:
	  cpswaps(from->u.clientMessage.u.s.shorts0,
	     to->u.clientMessage.u.s.shorts0);
	  cpswaps(from->u.clientMessage.u.s.shorts1,
	     to->u.clientMessage.u.s.shorts1);
	  cpswaps(from->u.clientMessage.u.s.shorts2,
	     to->u.clientMessage.u.s.shorts2);
	  cpswaps(from->u.clientMessage.u.s.shorts3,
	     to->u.clientMessage.u.s.shorts3);
	  cpswaps(from->u.clientMessage.u.s.shorts4,
	     to->u.clientMessage.u.s.shorts4);
	  cpswaps(from->u.clientMessage.u.s.shorts5,
	     to->u.clientMessage.u.s.shorts5);
	  cpswaps(from->u.clientMessage.u.s.shorts6,
	     to->u.clientMessage.u.s.shorts6);
	  cpswaps(from->u.clientMessage.u.s.shorts7,
	     to->u.clientMessage.u.s.shorts7);
	  cpswaps(from->u.clientMessage.u.s.shorts8,
	     to->u.clientMessage.u.s.shorts8);
	  cpswaps(from->u.clientMessage.u.s.shorts9,
	     to->u.clientMessage.u.s.shorts9);
	  break;
       case 32:
	  cpswapl(from->u.clientMessage.u.l.longs0,
	     to->u.clientMessage.u.l.longs0);
	  cpswapl(from->u.clientMessage.u.l.longs1,
	     to->u.clientMessage.u.l.longs1);
	  cpswapl(from->u.clientMessage.u.l.longs2,
	     to->u.clientMessage.u.l.longs2);
	  cpswapl(from->u.clientMessage.u.l.longs3,
	     to->u.clientMessage.u.l.longs3);
	  cpswapl(from->u.clientMessage.u.l.longs4,
	     to->u.clientMessage.u.l.longs4);
	  break;
       }
}

void
SKeymapNotifyEvent(from, to)
    xEvent	*from, *to;
{
    /* Keymap notify events are special; they have no
       sequence number field, and contain entirely 8-bit data */
    *to = *from;
}

void
WriteSConnectionInfo(pClient, size, pInfo)
    ClientPtr		pClient;
    unsigned long	size;
    char 		*pInfo;
{
    int		i, j, k;
#if MAXSCREENS != 1
    ScreenPtr	pScreen;
#endif
    DepthPtr	pDepth;
    char	*pInfoT, *pInfoTBase;
    xConnSetup	*pConnSetup = (xConnSetup *)pInfo;

    pInfoT = pInfoTBase = (char *) ALLOCATE_LOCAL((int)size);
    if (!pInfoTBase)
    {
	pClient->noClientException = -1;
	return;
    }
    SwapConnSetup(pConnSetup, (xConnSetup *)pInfoT);
    pInfo += sizeof(xConnSetup);
    pInfoT += sizeof(xConnSetup);

    /* Copy the vendor string */
    i = (pConnSetup->nbytesVendor + 3) & ~3;
    bcopy(pInfo, pInfoT, i);
    pInfo += i;
    pInfoT += i;

    /* The Pixmap formats don't need to be swapped, just copied. */
    i = sizeof(xPixmapFormat) * screenInfo.numPixmapFormats;
    bcopy(pInfo, pInfoT, i);
    pInfo += i;
    pInfoT += i;

#if MAXSCREENS != 1
    for(i = 0; i < screenInfo.numScreens; i++)
#endif
    {
#if MAXSCREENS != 1
	pScreen = screenInfo.screens[i];
#endif
	SwapWinRoot((xWindowRoot *)pInfo, (xWindowRoot *)pInfoT);
	pInfo += sizeof(xWindowRoot);
	pInfoT += sizeof(xWindowRoot);
#if MAXSCREENS == 1
	pDepth = screenInfo.screen.allowedDepths;
	for(j = 0; j < screenInfo.screen.numDepths; j++, pDepth++)
#else
	pDepth = pScreen->allowedDepths;
	for(j = 0; j < pScreen->numDepths; j++, pDepth++)
#endif
	{
            ((xDepth *)pInfoT)->depth = ((xDepth *)pInfo)->depth;
	    cpswaps(((xDepth *)pInfo)->nVisuals, ((xDepth *)pInfoT)->nVisuals);
	    pInfo += sizeof(xDepth);
	    pInfoT += sizeof(xDepth);
	    for(k = 0; k < pDepth->numVids; k++)
	    {
		SwapVisual((xVisualType *)pInfo, (xVisualType *)pInfoT);
		pInfo += sizeof(xVisualType);
		pInfoT += sizeof(xVisualType);
	    }
	}
    }
    (void)WriteToClient(pClient, (int)size, (char *) pInfoTBase);
    DEALLOCATE_LOCAL(pInfoTBase);
}

void
SwapConnSetup(pConnSetup, pConnSetupT)
    xConnSetup 	*pConnSetup, *pConnSetupT;
{
    cpswapl(pConnSetup->release, pConnSetupT->release);
    cpswapl(pConnSetup->ridBase, pConnSetupT->ridBase);
    cpswapl(pConnSetup->ridMask, pConnSetupT->ridMask);
    cpswapl(pConnSetup->motionBufferSize, pConnSetupT->motionBufferSize);
    cpswaps(pConnSetup->nbytesVendor, pConnSetupT->nbytesVendor);
    cpswaps(pConnSetup->maxRequestSize, pConnSetupT->maxRequestSize);
    pConnSetupT->minKeyCode = pConnSetup->minKeyCode;
    pConnSetupT->maxKeyCode = pConnSetup->maxKeyCode;
    pConnSetupT->numRoots = pConnSetup->numRoots;
    pConnSetupT->numFormats = pConnSetup->numFormats;
    pConnSetupT->imageByteOrder = pConnSetup->imageByteOrder;
    pConnSetupT->bitmapBitOrder = pConnSetup->bitmapBitOrder;
    pConnSetupT->bitmapScanlineUnit = pConnSetup->bitmapScanlineUnit;
    pConnSetupT->bitmapScanlinePad = pConnSetup->bitmapScanlinePad;
}

void
SwapWinRoot(pRoot, pRootT)
    xWindowRoot	*pRoot, *pRootT;
{
    cpswapl(pRoot->windowId, pRootT->windowId);
    cpswapl(pRoot->defaultColormap, pRootT->defaultColormap);
    cpswapl(pRoot->whitePixel, pRootT->whitePixel);
    cpswapl(pRoot->blackPixel, pRootT->blackPixel);
    cpswapl(pRoot->currentInputMask, pRootT->currentInputMask);
    cpswaps(pRoot->pixWidth, pRootT->pixWidth);
    cpswaps(pRoot->pixHeight, pRootT->pixHeight);
    cpswaps(pRoot->mmWidth, pRootT->mmWidth);
    cpswaps(pRoot->mmHeight, pRootT->mmHeight);
    cpswaps(pRoot->minInstalledMaps, pRootT->minInstalledMaps);
    cpswaps(pRoot->maxInstalledMaps, pRootT->maxInstalledMaps);
    cpswapl(pRoot->rootVisualID, pRootT->rootVisualID);
    pRootT->backingStore = pRoot->backingStore;
    pRootT->saveUnders = pRoot->saveUnders;
    pRootT->rootDepth = pRoot->rootDepth;
    pRootT->nDepths = pRoot->nDepths;
}

void
SwapVisual(pVis, pVisT)
    xVisualType 	*pVis, *pVisT;
{
    cpswapl(pVis->visualID, pVisT->visualID);
    pVisT->class = pVis->class;
    pVisT->bitsPerRGB = pVis->bitsPerRGB;
    cpswaps(pVis->colormapEntries, pVisT->colormapEntries);
    cpswapl(pVis->redMask, pVisT->redMask);
    cpswapl(pVis->greenMask, pVisT->greenMask);
    cpswapl(pVis->blueMask, pVisT->blueMask);
}

void
WriteSConnSetupPrefix(pClient, pcsp)
    ClientPtr		pClient;
    xConnSetupPrefix	*pcsp;
{
    xConnSetupPrefix	cspT;

    cspT.success = pcsp->success;
    cspT.lengthReason = pcsp->lengthReason;
    cpswaps(pcsp->majorVersion, cspT.majorVersion);
    cpswaps(pcsp->minorVersion, cspT.minorVersion);
    cpswaps(pcsp->length, cspT.length);
    (void)WriteToClient(pClient, sizeof(cspT), (char *) &cspT);
}
